import asyncio
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from typing import Any
from typing import Awaitable
from typing import List
from typing import Union
from typing import cast

from aflowey import AsyncFlow
from aflowey import aflow
from aflowey.runner import AsyncRunner, ExecutorType
from aflowey.single_executor import SingleFlowExecutor
from aflowey.types import AnyCallable

FlowOrListFlow = Union[List[AsyncFlow], AsyncFlow]


class AsyncFlowExecutor:
    """
    Execute several flows concurrently

    >>> await (aexec().from_flows(flows) | flow).run()

    """

    def __init__(self) -> None:
        """
        Creates a new async flow executor
        """
        self.flows: List[Union[AsyncFlow, List[AsyncFlow]]] = []

    def __or__(self, flow: Any) -> "AsyncFlowExecutor":
        """add a flow to execute in parallel"""
        return self.from_flows(flow)

    @staticmethod
    async def _execute_one_flow(flow: AsyncFlow, **kwargs: Any) -> Any:
        """Run"""
        return await SingleFlowExecutor(flow).execute_flow(is_root=True, **kwargs)

    def _execute_or_gather(self, flow: FlowOrListFlow, **kwargs: Any) -> Awaitable[Any]:
        if isinstance(flow, list):
            flows_task = [self._execute_one_flow(flow, **kwargs) for flow in flow]
            return asyncio.gather(*flows_task, **kwargs)
        return self._execute_one_flow(flow, **kwargs)

    async def run(self, **kwargs: Any) -> Any:
        """main function to run stuff in parallel"""
        # if other flow run in parallel
        flows = [self._execute_or_gather(flow, **kwargs) for flow in self.flows]
        return await asyncio.gather(*flows, **kwargs)

    def thread_runner(self, **kwargs) -> AsyncRunner:
        return AsyncRunner(self.run, ExecutorType.THREAD_POOL, **kwargs)

    def process_runner(self, **kwargs) -> AsyncRunner:
        return AsyncRunner(self.run, ExecutorType.PROCESS_POOL, **kwargs)

    def with_thread_runner(
        self, thread_pool: ThreadPoolExecutor, **kwargs
    ) -> AsyncRunner:
        return AsyncRunner(self.run, thread_pool, **kwargs)

    def with_process_runner(
        self, process_pool: ProcessPoolExecutor, **kwargs
    ) -> AsyncRunner:
        return AsyncRunner(self.run, process_pool, **kwargs)

    @staticmethod
    def ensure_flow(fn: Any, arg: Any = None) -> AsyncFlow:
        if isinstance(fn, AsyncFlow):
            return fn
        flow = aflow.empty() if arg is None else aflow.from_args(arg)
        return cast(AsyncFlow, flow >> fn)

    def from_flows(self, flows: Any) -> "AsyncFlowExecutor":
        """create a new executor from one flow or array of flows"""
        if not isinstance(flows, list):
            flows = AsyncFlowExecutor.ensure_flow(flows)
            self.flows.append(flows)
        else:
            new_flows = [AsyncFlowExecutor.ensure_flow(value) for value in flows]
            self.flows.append(new_flows)
        return self

    @staticmethod
    def empty() -> "AsyncFlowExecutor":
        return AsyncFlowExecutor()

    @staticmethod
    def starmap(*flows: Any) -> AnyCallable:
        async def wrapper(arg: Any) -> Any:
            new_flows = [AsyncFlowExecutor.ensure_flow(fn, arg) for fn in flows]
            result = await AsyncFlowExecutor().from_flows(new_flows).run()
            return result[0]

        return wrapper


async_exec = aexec = AsyncFlowExecutor
flows_from_arg = spawn_flows = run_flows = astarmap = AsyncFlowExecutor.starmap



from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from enum import Enum, auto
from typing import Union, Any, Mapping, Optional

from aflowey.context import Context, ctx_var, executor_var
from aflowey.types import Opt, Executor


class ExecutorType(Enum):
    THREAD_POOL = auto()
    PROCESS_POOL = auto()


def init_executor_if_needed(
    executor: Opt[Union[Executor, ExecutorType]], **kwargs: Any
) -> Opt[Executor]:
    if executor is None:  # pragma: no cover
        return None
    if isinstance(executor, ExecutorType):
        if executor is ExecutorType.THREAD_POOL:
            return ThreadPoolExecutor(**kwargs)
        elif executor is ExecutorType.PROCESS_POOL:
            return ProcessPoolExecutor(**kwargs)
        raise ValueError("Wrong provided executor type")  # pragma: no cover
    return executor


class AsyncRunner:
    """
    Runner which holds execution options

    >>> async_exec = aexec() | flow1 | flow2 | [flow3, flow4]
    >>> with async_exec.thread_runner(max_workers=4) as runner:
    >>>     a, b, (c, d) = await runner.run(async_exec)
    """

    def __init__(
        self, func, executor: Union[Executor, Opt[ExecutorType]] = None, **kwargs
    ):
        self.func = func
        self._executor = init_executor_if_needed(executor, **kwargs)
        self._result: Optional[Any] = None
        self._ctx = None
        self._token = None

    def __enter__(self) -> "AsyncRunner":
        """if no executor provided, raise an error as the use of the with
        keyword is useless. Creates the context of the current executor"""
        # do not override it
        if self._executor is None:
            raise ValueError(
                "Trying to use with context with not executor provided"
            )  # pragma: no cover
        self._executor.__enter__()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """closing the current executor"""
        if self._executor is None:
            raise ValueError(
                "Trying to use with context with not executor provided"
            )  # pragma: no cover
        self._executor.__exit__(exc_type, exc_val, exc_tb)

    async def run(self, context: Opt[Union[Context, Mapping]] = None, **kwargs):
        """all kwargs stuff will be passed to the executor

        Args:
            context: context var variable
            kwargs: passed to the gather function which run coroutine simultaneously
        """
        _ctx = context or {}
        _ctx = _ctx if isinstance(_ctx, Context) else Context(**_ctx)

        self._token = ctx_var.set(_ctx)
        executor_var.set(self._executor)
        # run the function of the executor
        self._result = await self.func(**kwargs)

        # reset to avoid memory leak
        ctx_var.reset(self._token)
        return self._result



from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import ThreadPoolExecutor
from typing import Any
from typing import Callable
from typing import Coroutine
from typing import Optional
from typing import TypeVar
from typing import Union


Opt = Optional
Function = TypeVar("Function", bound=Callable[..., Any])
Executor = Optional[Union[ThreadPoolExecutor, ProcessPoolExecutor]]
AnyCoroutineFunction = Callable[[Any, Any], Coroutine[Any, Any, Any]]
AnyCallable = Callable[[Any], Any]



from aflowey.async_flow import aflow
from aflowey.async_flow import async_flow
from aflowey.async_flow import AsyncFlow
from aflowey.executor import aexec
from aflowey.executor import astarmap
from aflowey.executor import async_exec
from aflowey.executor import AsyncFlowExecutor
from aflowey.executor import flows_from_arg
from aflowey.executor import run_flows
from aflowey.executor import spawn_flows
from aflowey.f import (
    F,
)
from aflowey.functions import apartial
from aflowey.functions import breaker
from aflowey.functions import ensure_callable
from aflowey.functions import erratic
from aflowey.functions import F0
from aflowey.functions import f0
from aflowey.functions import F1
from aflowey.functions import f1
from aflowey.functions import flog
from aflowey.functions import identity
from aflowey.functions import imp
from aflowey.functions import impure
from aflowey.functions import lift
from aflowey.functions import log
from aflowey.functions import may_fail
from aflowey.functions import partial
from aflowey.functions import side_effect
from aflowey.functions import spread
from aflowey.functions import spread_args
from aflowey.functions import spread_kw
from aflowey.functions import spread_kwargs
from aflowey.single_executor import (
    CANCEL_FLOW,
)

__all__ = [
    "F",
    "ensure_callable",
    "side_effect",
    "f0",
    "F0",
    "f1",
    "F1",
    "erratic",
    "breaker",
    "may_fail",
    "partial",
    "imp",
    "impure",
    "AsyncFlow",
    "aflow",
    "async_flow",
    "AsyncFlowExecutor",
    "async_exec",
    "aexec",
    "apartial",
    "partial",
    "spread",
    "spread_args",
    "spread_kw",
    "spread_kwargs",
    "identity",
    "CANCEL_FLOW",
    "flog",
    "log",
    "run_flows",
    "flows_from_arg",
    "spawn_flows",
    "astarmap",
    "lift",
]



from copy import copy
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from rich.tree import Tree

from aflowey.functions import ensure_callable
from aflowey.f import F
from aflowey.functions import ensure_f
from aflowey.functions import named
from aflowey.functions import side_effect
from aflowey.types import Function


class AsyncFlow:
    """
    Describe an async flow chaining function

    >>>flow = (AsyncFlow() >> gen1 >> gen2 >> gen3)

    >>>await flow.run()
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        # first arguments input
        self.args = args
        self.kwargs = kwargs

        # function to be executed in the flow
        self.aws: List[F] = []

        # intermediary store results
        self.steps: Dict[str, Any] = {}

        # attribute set by the runner
        self.executed: bool = False
        self.is_success: Optional[bool] = None

    def __call__(self, *args, **kwargs):
        return self

    @staticmethod
    def ensure_f_function(func: Union[Function, F]) -> F:
        new_func: Function = ensure_callable(func)
        return ensure_f(new_func)

    def __rshift__(
        self, aws: Union[List[Union[F, Function]], Union[F, Function], Any]
    ) -> "AsyncFlow":
        """add a new step to the flow

        Args:
            aws: list of callable or callable
        """
        if isinstance(aws, list):
            self.aws += [self.ensure_f_function(a) for a in aws]
        else:
            self.aws.append(self.ensure_f_function(aws))

        return self

    def __copy__(self) -> "AsyncFlow":
        """make a shallow copy"""
        aws = self.aws[:]
        args = self.args[:]
        kwargs = self.kwargs.copy()
        new_flow = self.__class__(*args, **kwargs)
        new_flow.aws = aws
        return new_flow

    async def run(self, **kwargs) -> Any:
        """run the flow

        Returns:
            coroutine
        """
        from aflowey.single_executor import SingleFlowExecutor

        return await SingleFlowExecutor(self).execute_flow(is_root=True, **kwargs)

    @staticmethod
    def from_args(*args: Any, **kwargs: Any) -> "AsyncFlow":
        """create a flow with given arguments as first input"""
        return AsyncFlow(*args, **kwargs)

    @staticmethod
    def empty() -> "AsyncFlow":
        """create an empty flow"""
        return AsyncFlow()

    @staticmethod
    def from_flow(flow: "AsyncFlow") -> "AsyncFlow":
        """create a new flow from given flow, copying it
        (args, kwargs, and aws functions)
        """
        assert isinstance(flow, AsyncFlow)
        return copy(flow)

    def _to_rich_tree(self) -> Tree:
        base_tree = Tree(
            ":open_file_folder: flow",
            style="bright_blue",
            highlight=True,
            guide_style="uu bright_blue",
        )
        for aws in self.aws:
            inner_func = aws.func
            if not isinstance(inner_func, AsyncFlow):
                f_name = inner_func.__name__
                if f_name == "__aflowey_wrapped":  # pragma: no cover
                    base_tree.add(str(inner_func()))
                else:
                    base_tree.add(f_name)
            else:
                base_tree.add(inner_func._to_rich_tree())
        return base_tree

    def display(self) -> None:
        root = self._to_rich_tree()
        from rich import print

        print(root)


# aliases
aflow = async_flow = AsyncFlow


def step(
    *func: Union[F, Function], name: Optional[str] = None, impure: bool = False
) -> Union[Function, List[F], F]:
    if len(func) > 1 and impure is False:
        raise ValueError(
            "Can not have several functions for impure method"
        )  # pragma: no cover
    if impure:
        new_func = side_effect(*func)
        return new_func
    named_func = func[0]
    return named(named_func, name) if name is not None else named_func



import asyncio
import functools
import inspect
from typing import Any
from typing import Callable
from typing import cast
from typing import Iterable
from typing import List
from typing import Optional
from typing import Union

from loguru import logger

from aflowey.f import F
from aflowey.types import Function


def wrapper_async(func: Function) -> F:
    """wrap a function into a coroutine function
    Args:
        func: callable
    Returns:
        F instance
    """

    @functools.wraps(func)
    async def wrapper(*a: Any, **kw: Any) -> Any:
        return await func(*a, **kw)

    return F(wrapper)


def log(log_str: str = "", print_arg: bool = False) -> Any:
    """utility function to log between steps, printing argument if needed"""

    async def wrapped(last_result: Any) -> Any:
        if log_str:
            logger.info(log_str)
        if print_arg:  # pragma: no cover
            logger.info(last_result)
        return last_result

    return wrapped


flog = log


def partial(func: Function, *args: Any, **kwargs: Any) -> F:
    """make a partial function of the given func and ensure
    it will work in an async context
    """
    return F(cast(Callable[[], Any], functools.partial(func, *args, **kwargs)))


p = apartial = partial


def f1(func: Function, extractor: Optional[Function] = None) -> F:
    """wraps a one argument function (with arity 1) and allows
    to add an extractor to work on the input argument.
    """

    @functools.wraps(func)
    def wrapped(arg1: Any) -> Any:
        value = arg1 if extractor is None else extractor(arg1)
        return func(value)

    wrapped_fn = (
        wrapper_async(wrapped) if asyncio.iscoroutinefunction(func) else F(wrapped)
    )
    wrapped_fn.__F1__ = True  # type: ignore

    return wrapped_fn


F1 = f1


def f0(func: Function) -> F:
    """create a new function from a 0 arity function (takes 0 args). The new
    function takes exactly one argument and does not pass it to the wrapped
    function. It allows using a 0 arity function in a flow relatively easily.
    """

    # noinspection PyUnusedLocal
    @functools.wraps(func)
    def wrapped(arg1: Any) -> Any:
        return func()

    wrapped_fn = (
        wrapper_async(wrapped) if asyncio.iscoroutinefunction(func) else F(wrapped)
    )
    wrapped_fn.__F0__ = True  # type: ignore

    return wrapped_fn


F0 = f0


def may_fail(func: Function) -> Function:
    """simply for readability"""
    return func


breaker = erratic = may_fail


def spread_args(func: Function) -> F:
    """create a function which takes an iterable of args
    and spread it into the given function"""

    @functools.wraps(func)
    def wrapped(args: Iterable[Any]) -> Any:
        # if too much args, slice the args to given length
        return func(*args)

    wrapped_fn = (
        wrapper_async(wrapped) if asyncio.iscoroutinefunction(func) else F(wrapped)
    )
    return wrapped_fn


spread = spread_args


def spread_kwargs(func: Function) -> F:
    """create a function which takes a mapping of kwargs
    and spread it into the given function"""

    @functools.wraps(func)
    def wrapped(**kwargs: Any) -> Any:
        arg_spec = inspect.getfullargspec(func)
        args = set(arg_spec.args + arg_spec.kwonlyargs) - {"self"}
        new_kwargs = {key: kwargs[key] for key in args}
        return func(**new_kwargs)

    wrapped_fn = (
        wrapper_async(wrapped) if asyncio.iscoroutinefunction(func) else F(wrapped)
    )
    return wrapped_fn


spread_kw = spread_kwargs


def lift(f: Function, lift_op: Function = map) -> F:  # type: ignore[assignment]
    return p(lift_op, f)


def ensure_f(func: Function) -> F:
    """wrap the given function into an F instance"""
    if not isinstance(func, F):
        return F(func)
    return func


def make_impure(func: Union[Function, F]) -> F:
    """tags the given function as impure, i.e. consume an argument
    but does not create new one
    """
    # automatically create new function when the function
    # is a bound method and has no other input args
    func = ensure_f(func)
    func.__side_effect__ = True  # type: ignore
    return func


def side_effect(*func: Union[Function, F]) -> Union[List[F], F]:
    """
    take an array of function and tag it as side effects
    function
    """
    if len(func) == 1:
        return make_impure(func[0])
    return [make_impure(fu) for fu in func]


imp = impure = side_effect


def ensure_callable(x: Union[Any, Function]) -> Function:
    """ensure a given args is a callable by returning a new callable if not"""
    if not callable(x):

        # noinspection PyUnusedLocal
        def __aflowey_wrapped(*args: Any, **kwargs: Any) -> Any:
            return x

        return (
            cast(Function, __aflowey_wrapped)
            if not inspect.isawaitable(x)
            else wrapper_async(__aflowey_wrapped)
        )
    return cast(Function, x)


def named(func: Union[Function, F, Any], name: str) -> F:
    """tags a function as a named function"""
    if not callable(func):
        func = ensure_callable(func)
    func = ensure_f(func)
    func.__named__ = name  # type: ignore
    return func


def identity(x: Any) -> Any:
    return x


# helper functions
def is_f(func: Any) -> bool:  # pragma: no cover
    return isinstance(func, F)


def is_f0(func: Function) -> bool:
    return hasattr(func, "__F0__")


def is_f1(func: Function) -> bool:  # pragma: no cover
    return hasattr(func, "__F1__")


def is_side_effect(func: Function) -> bool:
    return hasattr(func, "__side_effect__")


def get_name(func: Function) -> str:
    if hasattr(func, "__named__"):
        return func.__named__  # type: ignore
    return ""



import asyncio
import functools
from concurrent.futures import ProcessPoolExecutor
from contextvars import copy_context
from inspect import isawaitable
from typing import Any
from typing import Awaitable
from typing import Union
import warnings

from loguru import logger

from aflowey.async_flow import AsyncFlow
from aflowey.context import executor_var
from aflowey.f import F
from aflowey.functions import get_name
from aflowey.functions import is_f0
from aflowey.functions import is_side_effect
from aflowey.types import Function


async def _exec(function: Union[F, Function], *a: Any, **kw: Any) -> Any:
    current_result = function(*a, **kw) if isinstance(function, F) else function

    while isawaitable(current_result):
        warn = kw.pop("_warn", None)
        if warn is not None:
            warnings.warn(
                "Function ran in a thread pool, and it seems that it should have not. "
                "It may be due to tweaks during generation of this higher order function. "
                "You may want to rewrite this function with simple async def definition."
            )
        current_result = await current_result
    if isinstance(current_result, F):
        # we launch a new execution only if we have an F
        # instance, occurring when function returning partial for example
        # and not on every callable (not especially wanted)
        return await _exec(current_result)
    return current_result


async def check_and_run_step(fn: F, *args: Any, **kwargs: Any) -> Awaitable[Any]:
    executor = executor_var.get()

    if fn.is_coroutine_function:
        return await _exec(fn, *args, **kwargs)

    new_fn = functools.partial(fn.func, *args, **kwargs)

    loop = asyncio.get_running_loop()

    # process executor does not have access to the context data
    if isinstance(executor, ProcessPoolExecutor):
        return await loop.run_in_executor(executor, new_fn)

    ctx = copy_context()
    func_call = functools.partial(ctx.run, new_fn)
    result = await loop.run_in_executor(executor, func_call)
    return await _exec(result, _warn=True)


class SingleFlowExecutor:
    """Single flow executor"""

    CANCEL_FLOW = object()

    def __init__(self, flow: AsyncFlow) -> None:
        self.flow = flow

    @staticmethod
    async def check_and_execute_flow_if_needed(
        maybe_flow: Union[Any, AsyncFlow], **kwargs: Any
    ) -> Any:
        """check if we have an async flow and execute it"""
        if isinstance(maybe_flow, AsyncFlow):
            return await SingleFlowExecutor(maybe_flow).execute_flow(
                is_root=False, **kwargs
            )
        return maybe_flow

    @staticmethod
    def need_to_cancel_flow(result: Any) -> bool:
        """check if we need to cancel flow checking sentinel"""
        if result is SingleFlowExecutor.CANCEL_FLOW:
            logger.debug("Received sentinel object, canceling flow...")
            return True
        return False

    @staticmethod
    def get_step_name(func: F, index: int) -> str:
        """get the step name"""
        return get_name(func) or str(index)

    def save_step(self, task: F, index: int, current_args: Any) -> None:
        """save step state in flow attribute"""
        step_name = self.get_step_name(task, index)
        self.flow.steps[step_name] = current_args

    def _check_current_args_if_side_effect(self, first_aws: F, res: Any) -> Any:
        if is_side_effect(first_aws):
            return self._get_result_from_early_abort()
            # self.flow.kwargs if self.flow.kwargs else self.flow.args
        return res

    async def _execute_first_step(self, first_aws: F, **kwargs: Any) -> Any:
        """executing the first step"""
        if not self.flow.args and not self.flow.kwargs and is_f0(first_aws):
            self.flow.args = (None,)

        res = await check_and_run_step(first_aws, *self.flow.args, **self.flow.kwargs)

        current_args = self._check_current_args_if_side_effect(first_aws, res)
        # if flow run it
        current_args = await self.check_and_execute_flow_if_needed(
            current_args, **kwargs
        )

        # memorize name
        self.save_step(first_aws, 0, current_args)
        return current_args

    def _get_result_from_early_abort(self):
        if self.flow.kwargs:
            return self.flow.kwargs
        if not self.flow.args:
            return None
        if self.flow.args and len(self.flow.args) == 1:
            return self.flow.args[0]
        return self.flow.args

    async def execute_flow(self, is_root: bool, **kwargs: Any) -> Any:
        """Main function to execute a flow"""
        if not self.flow.aws:
            return None

        # get first step
        try:
            first_aws = self.flow.aws[0]
            current_args = await self._execute_first_step(first_aws, **kwargs)

            if self.need_to_cancel_flow(current_args):
                # returning canceling flow
                if is_root:
                    return self._get_result_from_early_abort()
                return current_args

            for index, task in enumerate(self.flow.aws[1:]):

                result = await check_and_run_step(task, current_args)

                if is_side_effect(task):
                    # side effect task, does not return a value
                    self.save_step(task, index + 1, current_args)
                    if self.need_to_cancel_flow(result):
                        break  # pragma: no cover
                    continue  # pragma: no cover

                result = await self.check_and_execute_flow_if_needed(result, **kwargs)
                if self.need_to_cancel_flow(
                    result
                ):  # check if we need to cancel the flow
                    break

                current_args = result
                self.save_step(task, index + 1, current_args)
        except Exception as e:
            logger.error(e)
            return_exceptions: bool = kwargs.pop("return_exceptions", False)
            # setting flow information
            self.flow.is_success = False
            if return_exceptions is False:
                raise e
            return e
        else:
            self.flow.is_success = True
            return current_args
        finally:
            self.flow.executed = True


CANCEL_FLOW = SingleFlowExecutor.CANCEL_FLOW
